from __future__ import annotations

import copy
import itertools
from collections import defaultdict
from contextlib import contextmanager, suppress
from time import time
from typing import TYPE_CHECKING, Any, cast

from requests import Response

from streamlink.exceptions import PluginError, StreamError
from streamlink.logger import getLogger
from streamlink.stream.dash.manifest import MPD, freeze_timeline
from streamlink.stream.dash.segment import DASHSegment
from streamlink.stream.ffmpegmux import FFMPEGMuxer
from streamlink.stream.segmented import SegmentedStreamReader, SegmentedStreamWorker, SegmentedStreamWriter
from streamlink.stream.stream import Stream
from streamlink.utils.l10n import Language
from streamlink.utils.parse import parse_xml
from streamlink.utils.times import now


if TYPE_CHECKING:
    from collections.abc import Mapping
    from datetime import datetime

    from streamlink.session import Streamlink
    from streamlink.stream.dash.manifest import Representation


log = getLogger(".".join(__name__.split(".")[:-1]))


class DASHStreamWriter(SegmentedStreamWriter[DASHSegment, Response]):
    WRITE_CHUNK_SIZE: int = 8192

    reader: DASHStreamReader
    stream: DASHStream

    def fetch(self, segment: DASHSegment):
        if self.closed:
            return

        name = segment.name
        available_in = segment.available_in
        if available_in > 0:
            log.debug(f"{self.reader.mime_type} segment {name}: waiting {available_in:.01f}s ({segment.availability})")
            if not self.wait(available_in):
                log.debug(f"{self.reader.mime_type} segment {name}: cancelled")
                return
        log.debug(f"{self.reader.mime_type} segment {name}: downloading ({segment.availability})")

        request_args = copy.deepcopy(self.reader.stream.args)
        headers = request_args.pop("headers", {})

        if segment.byterange:
            start, length = segment.byterange
            end = str(start + length - 1) if length else ""
            headers["Range"] = f"bytes={start}-{end}"

        try:
            return self.session.http.get(
                segment.uri,
                timeout=self.timeout,
                exception=StreamError,
                headers=headers,
                retries=self.retries,
                **request_args,
            )
        except StreamError as err:
            log.error(f"{self.reader.mime_type} segment {name}: failed ({err})")

    def write(self, segment: DASHSegment, result: Response, *data):
        for chunk in result.iter_content(self.WRITE_CHUNK_SIZE):
            if self.closed:
                log.warning(f"{self.reader.mime_type} segment {segment.name}: aborted")
                return
            self.reader.buffer.write(chunk)

        log.debug(f"{self.reader.mime_type} segment {segment.name}: completed")


class DASHStreamWorker(SegmentedStreamWorker[DASHSegment, Response]):
    reader: DASHStreamReader
    writer: DASHStreamWriter
    stream: DASHStream

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mpd = self.stream.mpd

        self.manifest_reload_retries = self.session.options.get("dash-manifest-reload-attempts")
        self.duration_limit = self.stream.duration or self.duration_limit

    @contextmanager
    def sleeper(self, duration):
        """
        Do something and then wait for a given duration minus the time it took doing something
        """
        s = time()
        yield
        time_to_sleep = duration - (time() - s)
        if time_to_sleep > 0:
            self.wait(time_to_sleep)

    @property
    def _queue_deadline_wait(self) -> float:
        return self.mpd.minimumUpdatePeriod.total_seconds()

    def iter_segments(self):
        init = True
        back_off_factor = 1
        while not self.closed:
            # find the representation by ID
            representation = self.mpd.get_representation(self.reader.ident)

            if self.mpd.type == "static":
                refresh_wait = 5
            else:
                refresh_wait = (
                    max(
                        self.mpd.minimumUpdatePeriod.total_seconds(),
                        representation.period.duration.total_seconds() if representation else 0,
                    )
                    or 5
                )

            with self.sleeper(refresh_wait * back_off_factor):
                if not representation:
                    continue

                queued = False
                iter_segments = representation.segments(
                    sequence=self.sequence,
                    init=init,
                    # sync initial timeline generation between audio and video threads
                    timestamp=self.reader.timestamp if init else None,
                )
                for segment in iter_segments:
                    if init and not segment.init:
                        self.sequence = segment.num
                        init = False
                    queued |= yield segment

                # close worker if type is not dynamic (all segments were put into writer queue)
                if self.mpd.type != "dynamic":
                    self.close()
                    return

                # Implicit end of stream
                if self.check_queue_deadline(queued):
                    return

                if not self.reload():
                    back_off_factor = max(back_off_factor * 1.3, 10.0)
                else:
                    back_off_factor = 1

    def reload(self):
        if self.closed:
            return

        self.reader.buffer.wait_free()
        log.debug("Reloading manifest %r", self.reader.ident)
        res = self.session.http.get(
            cast("str", self.mpd.url),
            exception=StreamError,
            retries=self.manifest_reload_retries,
            **self.stream.args,
        )

        new_mpd = MPD(
            self.session.http.xml(res, ignore_ns=True),
            base_url=self.mpd.base_url,
            url=self.mpd.url,
            timelines=self.mpd.timelines,
        )

        new_rep = new_mpd.get_representation(self.reader.ident)
        if not new_rep:
            log.error(f"Failed to find matching DASH representation: {self.reader.ident!r}")
            self.close()
            return False

        with freeze_timeline(new_mpd):
            changed = len(list(itertools.islice(new_rep.segments(), 1))) > 0

        if changed:
            self.mpd = new_mpd

        return changed


class DASHStreamReader(SegmentedStreamReader[DASHSegment, Response]):
    __worker__ = DASHStreamWorker
    __writer__ = DASHStreamWriter

    worker: DASHStreamWorker
    writer: DASHStreamWriter
    stream: DASHStream

    def __init__(
        self,
        stream: DASHStream,
        representation: Representation,
        timestamp: datetime,
        name: str | None = None,
    ):
        super().__init__(stream, name=name)
        self.ident = representation.ident
        self.mime_type = representation.mimeType
        self.timestamp = timestamp


class DASHStream(Stream):
    """
    Implementation of the "Dynamic Adaptive Streaming over HTTP" protocol (MPEG-DASH)
    """

    __shortname__ = "dash"

    def __init__(
        self,
        session: Streamlink,
        mpd: MPD,
        video_representation: Representation | None = None,
        audio_representation: Representation | None = None,
        duration: float | None = None,
        **kwargs,
    ):
        """
        :param session: Streamlink session instance
        :param mpd: Parsed MPD manifest
        :param video_representation: Video representation
        :param audio_representation: Audio representation
        :param duration: Number of seconds until ending the stream
        :param kwargs: Additional keyword arguments passed to :meth:`requests.Session.request`
        """

        super().__init__(session)
        self.mpd = mpd
        self.video_representation = video_representation
        self.audio_representation = audio_representation
        self.duration = duration
        self.args = session.http.valid_request_args(**kwargs)

    def __json__(self):  # noqa: PLW3201
        json = dict(type=self.shortname())

        if self.mpd.url:
            args = self.args.copy()
            args.update(url=self.mpd.url)
            req = self.session.http.prepare_new_request(**args)
            json.update(
                # the MPD URL has already been prepared by the initial request in `parse_manifest`
                url=self.mpd.url,
                headers=dict(req.headers),
            )

        return json

    def to_url(self):
        if self.mpd.url is None:
            return super().to_url()

        # the MPD URL has already been prepared by the initial request in `parse_manifest`
        return self.mpd.url

    @staticmethod
    def fetch_manifest(session: Streamlink, url_or_manifest: str, **request_args) -> tuple[str, dict[str, Any]]:
        if url_or_manifest.startswith("<?xml"):
            return url_or_manifest, {}

        retries = session.options.get("dash-manifest-reload-attempts")
        args = session.http.valid_request_args(**request_args)
        res = session.http.get(url_or_manifest, retries=retries, **args)
        manifest: str = res.text
        url: str = res.url

        return manifest, dict(url=url, base_url=url)

    @staticmethod
    def parse_mpd(manifest: str, mpd_params: Mapping[str, Any]) -> MPD:
        node = parse_xml(manifest, ignore_ns=True)

        return MPD(node, **mpd_params)

    @classmethod
    def parse_manifest(
        cls,
        session: Streamlink,
        url_or_manifest: str,
        period: int | str = 0,
        with_video_only: bool = False,
        with_audio_only: bool = False,
        **kwargs,
    ) -> dict[str, DASHStream]:
        """
        Parse a DASH manifest file and return its streams.

        :param session: Streamlink session instance
        :param url_or_manifest: URL of the manifest file or an XML manifest string
        :param period: Which MPD period to use (index number (int) or ``id`` attribute (str)) for finding representations
        :param with_video_only: Also return video-only streams, otherwise only return muxed streams
        :param with_audio_only: Also return audio-only streams, otherwise only return muxed streams
        :param kwargs: Additional keyword arguments passed to :class:`DASHStream` or :meth:`requests.Session.request`
        """

        manifest, mpd_params = cls.fetch_manifest(session, url_or_manifest, **kwargs)
        passthrough_encrypted = session.options.get("stream-passthrough-encrypted")

        try:
            mpd = cls.parse_mpd(manifest, mpd_params)
        except Exception as err:
            raise PluginError(f"Failed to parse MPD manifest: {err}") from err

        source = mpd_params.get("url", "MPD manifest")
        video: list[Representation | None] = [None] if with_audio_only else []
        audio: list[Representation | None] = [None] if with_video_only else []

        available_periods = [f"{idx}{f' (id={p.id!r})' if p.id is not None else ''}" for idx, p in enumerate(mpd.periods)]
        log.debug(f"Available DASH periods: {', '.join(available_periods)}")

        try:
            if isinstance(period, int):
                period_selection = mpd.periods[period]
            else:
                period_selection = mpd.periods_map[period]
        except LookupError:
            raise PluginError(
                f"DASH period {period!r} not found. Select a valid period by index or by id attribute value.",
            ) from None

        # Search for suitable video and audio representations
        for aset in period_selection.adaptationSets:
            if aset.contentProtections and not passthrough_encrypted:
                raise PluginError(f"{source} is protected by DRM")
            for rep in aset.representations:
                if rep.contentProtections and not passthrough_encrypted:
                    raise PluginError(f"{source} is protected by DRM")
                if rep.mimeType.startswith("video"):
                    video.append(rep)
                elif rep.mimeType.startswith("audio"):  # pragma: no branch
                    audio.append(rep)

        if passthrough_encrypted:
            is_encrypted = any(
                aset.contentProtections or any(rep.contentProtections for rep in aset.representations)
                for aset in period_selection.adaptationSets
            )
            if is_encrypted:  # pragma: no branch
                log.warning(f"{source} is protected by DRM and won't be decrypted")

        if not video:
            video.append(None)
        if not audio:
            audio.append(None)

        locale = session.localization
        locale_lang = locale.language
        lang = None
        available_languages = set()

        # if the locale is explicitly set, prefer that language over others
        for aud in audio:
            if aud and aud.lang:
                available_languages.add(aud.lang)
                with suppress(LookupError):
                    if locale.explicit and aud.lang and Language.get(aud.lang) == locale_lang:
                        lang = aud.lang

        if not lang:
            # filter by the first language that appears
            lang = audio[0].lang if audio[0] else None

        log.debug(
            f"Available languages for DASH audio streams: {', '.join(available_languages) or 'NONE'} (using: {lang or 'n/a'})",
        )

        # if the language is given by the stream, filter out other languages that do not match
        if len(available_languages) > 1:
            audio = [a for a in audio if a and (a.lang is None or a.lang == lang)]

        ret = []
        for vid, aud in itertools.product(video, audio):
            if not vid and not aud:
                continue

            stream = DASHStream(session, mpd, vid, aud, **kwargs)
            stream_name = []

            if vid:
                stream_name.append(f"{vid.height or vid.bandwidth_rounded:0.0f}{'p' if vid.height else 'k'}")
            if aud and len(audio) > 1:
                stream_name.append(f"a{aud.bandwidth:0.0f}k")
            ret.append(("+".join(stream_name), stream))

        # rename duplicate streams
        dict_value_list = defaultdict(list)
        for k, v in ret:
            dict_value_list[k].append(v)

        def sortby_bandwidth(dash_stream: DASHStream) -> float:
            if dash_stream.video_representation:
                return dash_stream.video_representation.bandwidth
            if dash_stream.audio_representation:
                return dash_stream.audio_representation.bandwidth
            return 0  # pragma: no cover

        ret_new = {}
        for q in dict_value_list:
            items = dict_value_list[q]

            with suppress(AttributeError):
                items = sorted(items, key=sortby_bandwidth, reverse=True)

            for n in range(len(items)):
                if n == 0:
                    ret_new[q] = items[n]
                elif n == 1:
                    ret_new[f"{q}_alt"] = items[n]
                else:
                    ret_new[f"{q}_alt{n}"] = items[n]

        return ret_new

    def open(self):
        video, audio = None, None
        rep_video, rep_audio = self.video_representation, self.audio_representation

        timestamp = now()

        if rep_video:
            video = DASHStreamReader(self, rep_video, timestamp, name="video")
            log.debug("Opening DASH reader for: %r - %s", rep_video.ident, rep_video.mimeType)

        if rep_audio:
            audio = DASHStreamReader(self, rep_audio, timestamp, name="audio")
            log.debug("Opening DASH reader for: %r - %s", rep_audio.ident, rep_audio.mimeType)

        if video and audio and FFMPEGMuxer.is_usable(self.session):
            video.open()
            audio.open()
            return FFMPEGMuxer(self.session, video, audio, copyts=True).open()
        elif video:
            video.open()
            return video
        elif audio:
            audio.open()
            return audio



import abc


class UserInputRequester(abc.ABC):
    """
    Base Class / Interface for requesting user input

    e.g. from the console
    """

    @abc.abstractmethod
    def ask(self, prompt: str) -> str:
        """
        Ask the user for a text input, the input is not sensitive
        and can be echoed to the user

        :param prompt: message to display when asking for the input
        :return: the value of the user input
        """
        raise NotImplementedError

    @abc.abstractmethod
    def ask_password(self, prompt: str) -> str:
        """
        Ask the user for a text input, the input _is_ sensitive
        and should be masked as the user gives the input

        :param prompt: message to display when asking for the input
        :return: the value of the user input
        """
        raise NotImplementedError



"""
$description Japanese live radio simulcasts and time-shifted broadcasts for over 100 stations.
$url radiko.jp
$type live, vod
$region Japan
"""

import base64
import datetime
import hashlib
import random
import re
from urllib.parse import urlencode

from lxml.etree import XML

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.stream.hls import HLSStream


@pluginmatcher(
    re.compile(r"https?://radiko\.jp/(#!/)?(?P<state>live|ts)/(?P<station_id>[a-zA-Z0-9-]+)/?(?P<start_at>\d+)?"),
)
class Radiko(Plugin):
    _api_auth_1 = "https://radiko.jp/v2/api/auth1"
    _api_auth_2 = "https://radiko.jp/v2/api/auth2"
    _auth_key = "bcd151073c03b352e1ef2fd66c32209da9ca0afa"

    def _get_streams(self):
        state = self.match.group("state")
        station_id = self.match.group("station_id").upper()
        if state == "live":
            url, token = self._live(station_id)
        else:
            start_at = self.match.group("start_at")
            url, token = self._timefree(station_id, start_at)
        self.session.http.headers.update({
            "X-Radiko-AuthToken": token,
        })
        yield from HLSStream.parse_variant_playlist(self.session, url).items()

    def _live(self, station_id):
        live_url = f"http://f-radiko.smartstream.ne.jp/{station_id}/_definst_/simul-stream.stream/playlist.m3u8"
        token, _area_id = self._authorize()
        lsid = hashlib.md5(str(random.random()).encode("utf-8")).hexdigest()
        live_params = {
            "station_id": station_id,
            "l": 15,
            "lsid": lsid,
            "type": "b",
        }
        url = f"{live_url}?{urlencode(live_params)}"
        return url, token

    def _timefree(self, station_id, start_at):
        m3u8_url = "https://tf-rpaa.smartstream.ne.jp/tf/playlist.m3u8"
        token, _area_id = self._authorize()
        lsid = hashlib.md5(str(random.random()).encode("utf-8")).hexdigest()
        end_at = self._get_xml(start_at, station_id)
        m3u8_params = {
            "station_id": station_id,
            "start_at": start_at,
            "ft": start_at,
            "end_at": end_at,
            "to": end_at,
            "l": 15,
            "lsid": lsid,
            "type": "b",
        }
        url = f"{m3u8_url}?{urlencode(m3u8_params)}"
        return url, token

    def _authorize(self):
        headers = {
            "x-radiko-app": "pc_html5",
            "x-radiko-app-version": "0.0.1",
            "x-radiko-device": "pc",
            "x-radiko-user": "dummy_user",
        }
        self.session.http.headers.update(headers)
        r = self.session.http.get(self._api_auth_1)
        token = r.headers.get("x-radiko-authtoken")
        offset = int(r.headers.get("x-radiko-keyoffset"))
        length = int(r.headers.get("x-radiko-keylength"))
        partial_key = base64.b64encode(self._auth_key[offset : offset + length].encode("ascii")).decode("utf-8")
        headers = {
            "x-radiko-authtoken": token,
            "x-radiko-device": "pc",
            "x-radiko-partialkey": partial_key,
            "x-radiko-user": "dummy_user",
        }
        self.session.http.headers.update(headers)
        r = self.session.http.get(self._api_auth_2)
        if r.status_code == 200:
            return token, r.text.split(",")[0]

    def _get_xml(self, start_at, station_id):
        today = datetime.date(int(start_at[:4]), int(start_at[4:6]), int(start_at[6:8]))
        yesterday = today - datetime.timedelta(days=1)
        if int(start_at[8:10]) < 5:
            date = yesterday.strftime("%Y%m%d")
        else:
            date = today.strftime("%Y%m%d")
        api = f"http://radiko.jp/v3/program/station/date/{date}/{station_id}.xml"
        r = self.session.http.get(api)
        tree = XML(r.content)
        for x in tree[2][0][1].findall("prog"):
            if x.attrib["ft"] == start_at:
                return x.attrib["to"]


__plugin__ = Radiko



"""
$description Chinese video sharing website based in Shanghai, themed around animation, comics, and games (ACG).
$url live.bilibili.com
$type live
"""

import re

from streamlink.exceptions import NoStreamsError
from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://live\.bilibili\.com/(?P<channel>[^/?#]+)"),
)
class Bilibili(Plugin):
    _URL_API_V1_PLAYURL = "https://api.live.bilibili.com/room/v1/Room/playUrl"
    _URL_API_V2_PLAYINFO = "https://api.live.bilibili.com/xlive/web-room/v2/index/getRoomPlayInfo"

    SHOW_STATUS_OFFLINE = 0
    SHOW_STATUS_ONLINE = 1
    SHOW_STATUS_ROUND = 2

    @classmethod
    def stream_weight(cls, stream: str) -> tuple[float, str]:
        offset = 1 if "_alt" in stream else 0
        if stream.startswith("httpstream"):
            return 4 - offset, stream
        if stream.startswith("hls"):
            return 2 - offset, stream
        return super().stream_weight(stream)

    def _get_api_v1_playurl(self, room_id):
        return self.session.http.get(
            self._URL_API_V1_PLAYURL,
            params={
                "cid": room_id,
                "platform": "web",
                "quality": "4",
            },
            schema=validate.Schema(
                validate.parse_json(),
                validate.any(
                    validate.all(
                        {
                            "code": 0,
                            "data": {
                                "durl": [
                                    validate.all(
                                        {
                                            "url": validate.url(),
                                        },
                                        validate.get("url"),
                                    ),
                                ],
                            },
                        },
                        validate.get(("data", "durl")),
                    ),
                    validate.all(
                        {
                            "code": int,
                        },
                        validate.transform(lambda _: []),
                    ),
                ),
            ),
        )

    @property
    def _schema_v2_streams(self):
        return validate.all(
            [
                {
                    "protocol_name": str,
                    "format": validate.all(
                        [
                            {
                                "format_name": str,
                                "codec": validate.all(
                                    [
                                        {
                                            "codec_name": str,
                                            "base_url": str,
                                            "url_info": [
                                                {
                                                    "host": validate.url(),
                                                    "extra": str,
                                                },
                                            ],
                                        },
                                    ],
                                    validate.filter(lambda item: item["codec_name"] == "avc"),
                                ),
                            },
                        ],
                        validate.filter(lambda item: item["format_name"] in ("fmp4", "ts")),
                    ),
                },
            ],
            validate.filter(lambda item: item["protocol_name"] == "http_hls"),
        )

    def _get_api_v2_playinfo(self, room_id):
        return self.session.http.get(
            self._URL_API_V2_PLAYINFO,
            params={
                "room_id": room_id,
                "no_playurl": 0,
                "mask": 1,
                "qn": 0,
                "platform": "web",
                "protocol": "0,1",
                "format": "0,1,2",
                "codec": "0,1,2",
                "dolby": 5,
                "panorama": 1,
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "code": 0,
                    "data": {
                        "playurl_info": validate.none_or_all(
                            {
                                "playurl": {
                                    "stream": self._schema_v2_streams,
                                },
                            },
                            validate.get(("playurl", "stream")),
                        ),
                    },
                },
                validate.get(("data", "playurl_info")),
            ),
        )

    def _get_page_playinfo(self):
        data = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//script[contains(text(),'window.__NEPTUNE_IS_MY_WAIFU__={')][1]/text()"),
                validate.none_or_all(
                    validate.transform(str.replace, "window.__NEPTUNE_IS_MY_WAIFU__=", ""),
                    validate.parse_json(),
                    {
                        "roomInitRes": {
                            "data": {
                                "live_status": int,
                                "playurl_info": validate.none_or_all(
                                    {
                                        "playurl": {
                                            "stream": self._schema_v2_streams,
                                        },
                                    },
                                    validate.get(("playurl", "stream")),
                                ),
                            },
                        },
                        "roomInfoRes": {
                            "data": {
                                "room_info": {
                                    "live_id": int,
                                    "title": str,
                                    "area_name": str,
                                },
                                "anchor_info": {
                                    "base_info": {
                                        "uname": str,
                                    },
                                },
                            },
                        },
                    },
                    validate.union_get(
                        ("roomInfoRes", "data", "room_info", "live_id"),
                        ("roomInfoRes", "data", "anchor_info", "base_info", "uname"),
                        ("roomInfoRes", "data", "room_info", "area_name"),
                        ("roomInfoRes", "data", "room_info", "title"),
                        ("roomInitRes", "data", "live_status"),
                        ("roomInitRes", "data", "playurl_info"),
                    ),
                ),
            ),
        )
        if not data:
            return

        self.id, self.author, self.category, self.title, live_status, streams = data
        if live_status != self.SHOW_STATUS_ONLINE:
            log.info("Channel is offline")
            raise NoStreamsError

        return streams

    def _get_streams(self):
        http_streams = self._get_api_v1_playurl(self.match["channel"])
        for http_stream in http_streams:
            if self.session.http.head(http_stream, raise_for_status=False).status_code >= 400:
                continue
            yield "httpstream", HTTPStream(self.session, http_stream)

        hls_streams = self._get_page_playinfo()
        if not hls_streams:
            log.debug("Falling back to _get_api_v2_playinfo()")
            hls_streams = self._get_api_v2_playinfo(self.match["channel"])

        for hls_stream in hls_streams or []:
            for stream_format in hls_stream["format"]:
                for codec in stream_format["codec"]:
                    for url_info in codec["url_info"]:
                        url = f"{url_info['host']}{codec['base_url']}{url_info['extra']}"
                        yield "hls", HLSStream(self.session, url)


__plugin__ = Bilibili



"""
$description Internet portal in Estonia, Latvia, and Lithuania providing daily news, ranging from gardening to politics.
$url delfi.lt
$url delfi.ee
$url delfi.lv
$type vod
"""

import itertools
import re
from urllib.parse import urlparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.parse import parse_qsd
from streamlink.utils.url import update_scheme


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:[\w-]+\.)?delfi\.(?P<tld>lt|lv|ee)"),
)
class Delfi(Plugin):
    _api = {
        "lt": "https://g2.dcdn.lt/vfe/data.php",
        "lv": "https://g.delphi.lv/vfe/data.php",
        "ee": "https://g4.nh.ee/vfe/data.php",
    }

    def _get_streams_api(self, video_id):
        log.debug(f"Found video ID: {video_id}")

        tld = self.match.group("tld")
        try:
            data = self.session.http.get(
                self._api.get(tld, "lt"),
                params=dict(video_id=video_id),
                schema=validate.Schema(
                    validate.parse_json(),
                    {
                        "success": True,
                        "data": {
                            "versions": {
                                str: validate.all(
                                    [
                                        {
                                            "type": str,
                                            "src": str,
                                        },
                                    ],
                                    validate.filter(lambda item: item["type"] == "application/x-mpegurl"),
                                ),
                            },
                        },
                    },
                    validate.get(("data", "versions")),
                ),
            )
        except PluginError:
            log.error("Failed to get streams from API")
            return

        for stream in itertools.chain(*data.values()):
            src = update_scheme("https://", stream["src"], force=False)
            yield from HLSStream.parse_variant_playlist(self.session, src).items()

    def _get_streams_delfi(self, src):
        try:
            data = self.session.http.get(
                src,
                schema=validate.Schema(
                    validate.parse_html(),
                    validate.xml_xpath_string(".//script[contains(text(),'embedJs.setAttribute(')][1]/text()"),
                    validate.none_or_all(
                        re.compile(r"embedJs\.setAttribute\('src',\s*'(.+?)'"),
                        validate.none_or_all(
                            validate.get(1),
                            validate.transform(lambda url: parse_qsd(urlparse(url).fragment)),
                            {"stream": str},
                            validate.get("stream"),
                            validate.parse_json(),
                            {
                                "versions": [
                                    {
                                        "hls": str,
                                    },
                                ],
                            },
                            validate.get("versions"),
                        ),
                    ),
                ),
            )
        except PluginError:
            log.error("Failed to get streams from iframe")
            return

        for stream in data:
            src = update_scheme("https://", stream["hls"], force=False)
            yield from HLSStream.parse_variant_playlist(self.session, src).items()

    def _get_streams(self):
        root = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
            ),
        )

        video_id = root.xpath("string(.//div[@data-provider='dvideo'][@data-id][1]/@data-id)")
        if video_id:
            return self._get_streams_api(str(video_id))

        yt_id = root.xpath("string(.//script[contains(@src,'/yt.js')][@data-video]/@data-video)")
        if yt_id:
            return self.session.streams(f"https://www.youtube.com/watch?v={yt_id}")

        yt_iframe = root.xpath("string(.//iframe[starts-with(@src,'https://www.youtube.com/')][1]/@src)")
        if yt_iframe:
            return self.session.streams(str(yt_iframe))

        delfi = root.xpath("string(.//iframe[@name='delfi-stream'][@src][1]/@src)")
        if delfi:
            return self._get_streams_delfi(str(delfi))


__plugin__ = Delfi



"""
$description Turkish live TV channels from Dogus Group, including Euro Star, Star and NTV.
$url eurostartv.com.tr
$url kralmuzik.com.tr
$url ntv.com.tr
$url startv.com.tr
$type live
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


@pluginmatcher(re.compile(r"https?://(?:www\.)?eurostartv\.com\.tr/canli-izle"))
@pluginmatcher(re.compile(r"https?://(?:www\.)?kralmuzik\.com\.tr/tv/.+"))
@pluginmatcher(re.compile(r"https?://(?:www\.)?ntv\.com\.tr/canli-yayin/ntv"))
@pluginmatcher(re.compile(r"https?://(?:www\.)?startv\.com\.tr/canli-yayin"))
class Dogus(Plugin):
    _re_live_hls = re.compile(r"'(https?://[^']+/live/hls/[^']+)'")
    _re_yt_script = re.compile(r"youtube\.init\('([\w-]{11})'")

    def _get_streams(self):
        root = self.session.http.get(self.url, schema=validate.Schema(validate.parse_html()))

        # https://www.ntv.com.tr/canli-yayin/ntv?youtube=true
        yt_iframe = root.xpath("string(.//iframe[contains(@src,'youtube.com')][1]/@src)")
        # https://www.startv.com.tr/canli-yayin
        dm_iframe = root.xpath("string(.//iframe[contains(@src,'dailymotion.com')][1]/@src)")
        # https://www.kralmuzik.com.tr/tv/kral-tv
        # https://www.kralmuzik.com.tr/tv/kral-pop-tv
        yt_script = root.xpath("string(.//script[contains(text(), 'youtube.init')][1]/text())")
        if yt_script:
            m = self._re_yt_script.search(yt_script)
            if m:
                yt_iframe = f"https://www.youtube.com/watch?v={m.group(1)}"

        iframe = yt_iframe or dm_iframe
        if iframe:
            return self.session.streams(iframe)

        # http://eurostartv.com.tr/canli-izle
        dd_script = root.xpath("string(.//script[contains(text(), '/live/hls/')][1]/text())")
        if dd_script:
            m = self._re_live_hls.search(dd_script)
            if m:
                return HLSStream.parse_variant_playlist(self.session, m.group(1))


__plugin__ = Dogus



"""
$description Turkish live TV channel owned by Acun Medya Group.
$url tv8.com.tr
$type live
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream, HLSStreamReader, HLSStreamWriter


log = getLogger(__name__)


class TV8HLSStreamWriter(HLSStreamWriter):
    ad_re = re.compile(r"/ad/|/crea/")

    def should_filter_segment(self, segment):
        return self.ad_re.search(segment.uri) is not None or super().should_filter_segment(segment)


class TV8HLSStreamReader(HLSStreamReader):
    __writer__ = TV8HLSStreamWriter


class TV8HLSStream(HLSStream):
    __reader__ = TV8HLSStreamReader


@pluginmatcher(
    re.compile(r"https?://www\.tv8\.com\.tr/canli-yayin"),
)
class TV8(Plugin):
    def _get_streams(self):
        hls_url = self.session.http.get(
            self.url,
            schema=validate.Schema(
                re.compile(r"""var\s+videoUrl\s*=\s*(?P<q>["'])(?P<hls_url>https?://.*?\.m3u8.*?)(?P=q)"""),
                validate.any(None, validate.get("hls_url")),
            ),
        )
        if hls_url is not None:
            return TV8HLSStream.parse_variant_playlist(self.session, hls_url)


__plugin__ = TV8



"""
$description Indonesian & international live TV channels and video on-demand service. OTT service from Vidio.
$url vidio.com
$type live, vod
$metadata id
$metadata title
"""

import re
from uuid import uuid4

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?vidio\.com/.+"),
)
class Vidio(Plugin):
    tokens_url = "https://www.vidio.com/live/{id}/tokens"

    def _get_stream_token(self, stream_id, stream_type):
        return self.session.http.post(
            self.tokens_url.format(id=stream_id),
            params={"type": stream_type},
            headers={"Referer": self.url},
            cookies={
                "ahoy_visit": str(uuid4()),
                "ahoy_visitor": str(uuid4()),
            },
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "token": str,
                    "hls_url": validate.any("", validate.url()),
                    "dash_url": validate.any("", validate.url()),
                },
                validate.union_get(
                    "token",
                    "hls_url",
                    "dash_url",
                ),
            ),
        )

    def _get_stream_data(self):
        return self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_find(".//*[@data-video-id]"),
                validate.union((
                    validate.get("data-video-id"),
                    validate.get("data-video-title"),
                    validate.all(
                        validate.get("data-video-has-token"),
                        validate.transform(lambda val: val and val != "false"),
                    ),
                    validate.get("data-vjs-clip-hls-url"),
                    validate.get("data-vjs-clip-dash-url"),
                )),
            ),
        )

    def _get_streams(self):
        self.session.http.headers.update({
            "Origin": "https://www.vidio.com/",
            "Referer": self.url,
        })

        self.id, self.title, has_token, hls_url, dash_url = self._get_stream_data()
        log.debug(f"{self.id=} {has_token=}")

        params = {}
        if has_token:
            token, hls_url, dash_url = self._get_stream_token(self.id, "dash")
            log.trace("token=%s", token)
            params.update([param.split("=", 1) for param in (token.split("&") if token else [])])

        if hls_url:
            return HLSStream.parse_variant_playlist(self.session, hls_url, params=params)
        if dash_url:
            return DASHStream.parse_manifest(self.session, dash_url, params=params)


__plugin__ = Vidio



"""
$description Live TV channels and video on-demand service from NRK, a Norwegian public, state-owned broadcaster.
$url tv.nrk.no
$url radio.nrk.no
$type live, vod
$region Norway
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:tv|radio)\.nrk\.no/(program|direkte|serie|podkast)(?:/.+)?/([^/]+)"),
)
class NRK(Plugin):
    _URL_MANIFEST = "https://psapi.nrk.no/playback/manifest/{manifest_type}/{program_id}"
    _URL_METADATA = "https://psapi.nrk.no/playback/metadata/{manifest_type}/{program_id}?eea-portability=true"

    _MAP_MANIFEST_TYPE = {
        "direkte": "channel",
        "serie": "program",
        "program": "program",
        "podkast": "podcast",
    }

    def _get_metadata(self, manifest_type, program_id):
        url_metadata = self._URL_METADATA.format(manifest_type=manifest_type, program_id=program_id)
        non_playable = self.session.http.get(
            url_metadata,
            schema=validate.Schema(
                validate.parse_json(),
                {
                    validate.optional("nonPlayable"): validate.none_or_all(
                        {
                            "reason": str,
                            "endUserMessage": str,
                        },
                        validate.union_get("reason", "endUserMessage"),
                    ),
                },
                validate.get("nonPlayable"),
            ),
        )
        if non_playable:
            reason, end_user_message = non_playable
            log.error(f"Not playable: {reason} - {end_user_message or 'error'}")
            return False

        return True

    def _update_program_id(self, program_id):
        new_program_id = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//meta[@property='nrk:program-id']/@content"),
            ),
        )
        return new_program_id or program_id

    def _get_assets(self, manifest_type, program_id):
        return self.session.http.get(
            self._URL_MANIFEST.format(manifest_type=manifest_type, program_id=program_id),
            headers={"Accept": "application/vnd.nrk.psapi+json; version=9; player=tv-player; device=player-core"},
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "playable": {
                        "assets": [
                            validate.all(
                                {
                                    "url": validate.url(),
                                    "format": str,
                                },
                                validate.union_get("format", "url"),
                            ),
                        ],
                    },
                },
                validate.get(("playable", "assets")),
            ),
        )

    def _get_streams(self):
        program_type, program_id = self.match.groups()
        manifest_type = self._MAP_MANIFEST_TYPE.get(program_type)
        if manifest_type is None:
            log.error(f"Unknown program type '{program_type}'")
            return

        program_id = self._update_program_id(program_id)
        if self._get_metadata(manifest_type, program_id) is False:
            return

        assets = self._get_assets(manifest_type, program_id)

        # just return the first item
        for stream_type, stream_url in assets:
            if stream_type == "HLS":
                return HLSStream.parse_variant_playlist(self.session, stream_url)
            return [("live", HTTPStream(self.session, stream_url))]


__plugin__ = NRK



"""
$description Live TV channels from STV, a Scottish free-to-air broadcaster.
$url player.stv.tv
$type live
$metadata title
$region United Kingdom
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://player\.stv\.tv/live"),
)
class STV(Plugin):
    API_URL = "https://player.api.stv.tv/v1/streams/stv/"

    def get_title(self):
        if self.title is None:
            self._get_api_results()
        return self.title

    def _get_api_results(self):
        res = self.session.http.get(self.API_URL)
        data = self.session.http.json(res)

        if data["success"] is False:
            raise PluginError(data["reason"]["message"])

        try:
            self.title = data["results"]["now"]["title"]
        except KeyError:
            self.title = "STV"

        return data

    def _get_streams(self):
        hls_url = self._get_api_results()["results"]["streamUrl"]
        return HLSStream.parse_variant_playlist(self.session, hls_url)


__plugin__ = STV



"""
$description Russian live-streaming and video hosting social platform.
$url ok.ru
$url mobile.ok.ru
$type live, vod
$metadata id
$metadata author
$metadata title
"""

import re
from urllib.parse import unquote

from streamlink.exceptions import NoStreamsError
from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.dash import DASHStream
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream


log = getLogger(__name__)


@pluginmatcher(
    name="default",
    pattern=re.compile(r"https?://(?:www\.)?ok\.ru/"),
)
@pluginmatcher(
    name="mobile",
    pattern=re.compile(r"https?://m(?:obile)?\.ok\.ru/"),
)
class OKru(Plugin):
    QUALITY_WEIGHTS = {
        "full": 1080,
        "1080": 1080,
        "hd": 720,
        "720": 720,
        "sd": 480,
        "480": 480,
        "360": 360,
        "low": 360,
        "lowest": 240,
        "mobile": 144,
    }

    @classmethod
    def stream_weight(cls, stream: str) -> tuple[float, str]:
        weight = cls.QUALITY_WEIGHTS.get(stream)
        if weight:
            return weight, "okru"

        return super().stream_weight(stream)

    def _canonicalize_mobile_url(self):
        url = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//head/link[@rel='canonical'][@href][1]/@href"),
                validate.none_or_all(
                    validate.url(),
                ),
            ),
        )
        if not url or not self.matchers["default"].pattern.match(url):
            raise NoStreamsError

        self.url = url

    def _get_streams(self):
        if self.matches["mobile"]:
            self._canonicalize_mobile_url()

        schema_metadata = validate.Schema(
            validate.parse_json(),
            {
                validate.optional("author"): validate.all({validate.optional("name"): str}, validate.get("name")),
                validate.optional("movie"): validate.all(
                    {
                        validate.optional("id"): str,
                        validate.optional("title"): str,
                    },
                    validate.union_get("id", "title"),
                ),
                validate.optional("hlsManifestUrl"): validate.url(),
                validate.optional("hlsMasterPlaylistUrl"): validate.url(),
                validate.optional("liveDashManifestUrl"): validate.url(),
                validate.optional("videos"): [
                    validate.all(
                        {
                            "name": str,
                            "url": validate.url(),
                        },
                        validate.union_get("name", "url"),
                    ),
                ],
            },
        )

        metadata, metadata_url = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_find(".//*[@data-options]"),
                validate.get("data-options"),
                validate.parse_json(),
                {
                    "flashvars": {
                        validate.optional("metadata"): str,
                        validate.optional("metadataUrl"): validate.all(
                            validate.transform(unquote),
                            validate.url(),
                        ),
                    },
                },
                validate.get("flashvars"),
                validate.union_get("metadata", "metadataUrl"),
            ),
        )

        self.session.http.headers.update({"Referer": self.url})

        if not metadata and metadata_url:
            metadata = self.session.http.post(metadata_url).text

        log.trace("%r", metadata)

        data = schema_metadata.validate(metadata)

        self.author = data.get("author")
        self.id, self.title = data.get("movie", (None, None))

        for hls_url in data.get("hlsManifestUrl"), data.get("hlsMasterPlaylistUrl"):
            if hls_url is not None:
                return HLSStream.parse_variant_playlist(self.session, hls_url)

        if data.get("liveDashManifestUrl"):
            return DASHStream.parse_manifest(self.session, data.get("liveDashManifestUrl"))

        return {
            f"{self.QUALITY_WEIGHTS[name]}p" if name in self.QUALITY_WEIGHTS else name: HTTPStream(self.session, url)
            for name, url in data.get("videos", [])
        }


__plugin__ = OKru



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal


if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from re import Pattern


class SchemaContainer:
    """
    A simple schema container.
    """

    def __init__(self, schema):
        self.schema = schema


class _CollectionSchemaContainer(SchemaContainer):
    def __init__(self, *schemas):
        super().__init__(schemas)


class AllSchema(_CollectionSchemaContainer):
    """
    A collection of schemas where each schema must be valid.

    Validates one schema after another with the input value of the return value of the previous one.

    Example:

    .. code-block:: python

        # `validate.Schema` is a subclass of `AllSchema` (`validate.all`)
        schema = validate.Schema(
            int,
            validate.transform(lambda val: val + 1),
            lambda val: val < 3,
        )
        assert schema.validate(1) == 2
        schema.validate("a")  # raises ValidationError
        schema.validate(2)  # raises ValidationError

    :param Any \\*schemas: Schemas where each one must be valid
    :return: The return value of the last schema
    """


class AnySchema(_CollectionSchemaContainer):
    """
    A collection of schemas where at least one schema must be valid.

    Validates one schema after another with the same input value until the first one succeeds.

    Example:

    .. code-block:: python

        schema = validate.Schema(
            validate.any(int, float, str),
        )
        assert schema.validate(123) == 123
        assert schema.validate(123.0) == 123.0
        assert schema.validate("123") == "123"
        schema.validate(None)  # raises ValidationError

    :param Any \\*schemas: Schemas where at least one must be valid
    :raise ValidationError: Error collection of all schema validations if none succeeded
    :return: The return value of the first valid schema
    """


class NoneOrAllSchema(_CollectionSchemaContainer):
    """
    Similar to :class:`AllSchema`, but skips the validation if the input value is ``None``.

    This is useful for optional validation results, e.g. when validating a potential match of a regular expression.

    Example:

    .. code-block:: python

        schema = validate.Schema(
            validate.none_or_all(
                int,
                lambda val: val < 2,
            ),
        )
        assert schema.validate(None) is None
        assert schema.validate(1) == 1
        schema.validate("123")  # raises ValidationError
        schema.validate(2)  # raises ValidationError

    :param Any \\*schemas: Schemas where each one must be valid, unless the input is ``None``
    :raise ValidationError: Error wrapper of the failed schema validation
    :return: ``None`` if the input is ``None``, or the return value of the last schema
    """


class TransformSchema:
    """
    A transform function which receives the input value as the argument, with optional custom arguments and keywords.

    Example:

    .. code-block:: python

        schema = validate.Schema(
            validate.transform(lambda val: val + 1),
            validate.transform(operator.lt, 3),
        )
        assert schema.validate(1) is True
        assert schema.validate(2) is False

    :param func: A transform function
    :param \\*args: Additional arguments
    :param \\*\\*kwargs: Additional keywords
    :raise ValidationError: If the transform function is not callable
    :return: The return value of the transform function
    """

    def __init__(
        self,
        func: Callable,
        *args,
        **kwargs,
    ):
        self.func = func
        self.args = args
        self.kwargs = kwargs


class OptionalSchema:  # noqa: B903
    """
    An optional key set in a :class:`dict`.

    See the :func:`dict <streamlink.plugin.api.validate.validate_dict>` validation and the :class:`UnionSchema`.
    """

    def __init__(self, key: Any):
        self.key = key


class ListSchema(_CollectionSchemaContainer):
    """
    A list of schemas where every item must be valid, as well as the input type and length.

    Please note the difference between :class:`ListSchema`
    and the :func:`list <streamlink.plugin.api.validate.validate_sequence()>` validation.

    Example:

    .. code-block:: python

        schema = validate.Schema(
            validate.list(1, 2, int),
        )
        assert schema.validate([1, 2, 3]) == [1, 2, 3]
        schema.validate(None)  # raises ValidationError
        schema.validate([1, 2])  # raises ValidationError
        schema.validate([3, 2, 1])  # raises ValidationError

    :param Any \\*schema: Schemas where each one must be valid
    :raise ValidationError: If the input is not a :class:`list`
    :raise ValidationError: If the input's length is not equal to the number of schemas
    :return: A new :class:`list <builtins.list>` with the validated input
    """


class AttrSchema(SchemaContainer):
    """
    Validate attributes of an input object according to a :class:`dict`'s key-value pairs.

    Example:

    .. code-block:: python

        schema = validate.Schema(
            validate.attr({
                "a": str,
                "b": int,
            }),
        )
        assert schema.validate(obj) is not obj
        schema.validate(obj_without_a)  # raises ValidationError
        schema.validate(obj_b_is_str)  # raises ValidationError

    :param dict[str, Any] schema: A :class:`dict` with attribute validations
    :raise ValidationError: If the input doesn't have one of the schema's attributes
    :return: A copy of the input object with validated attributes
    """


class GetItemSchema:  # noqa: B903
    """
    Get an ``item`` from the input.

    The input can be anything that implements :func:`__getitem__()`,
    as well as :class:`lxml.etree.Element` objects where element attributes are looked up.

    Returns the ``default`` value if ``item`` was not found.

    Unless ``strict`` is set to ``True``, the ``item`` can be a :class:`tuple` of items for a recursive lookup.
    In this case, the ``default`` value is only returned if the leaf-input-object doesn't contain the current ``item``.

    Example:

    .. code-block:: python

        schema = validate.Schema(
            validate.get("name", default="unknown"),
        )
        assert schema.validate({"name": "user"}) == "user"
        assert schema.validate(re.match(r"Hello, (?P<name>\\w+)!", "Hello, user!")) == "user"
        assert schema.validate(lxml.etree.XML(\"\"\"<elem name="abc"/>\"\"\")) == "abc"
        assert schema.validate({}) == "unknown"
        schema.validate(None)  # raises ValidationError

    .. code-block:: python

        schema = validate.Schema(
            validate.get(("a", "b", "c")),
        )
        assert schema.validate({"a": {"b": {"c": "d"}}}) == "d"
        assert schema.validate({"a": {"b": {}}}) is None
        schema.validate({"a": {}})  # raises ValidationError

    :param item: The lookup key, or a :class:`tuple` of recursive lookup keys
    :param default: Optional custom default value
    :param strict: If ``True``, don't perform recursive lookups with the :class:`tuple` item
    :raise ValidationError: If the input doesn't implement :func:`__getitem__()`
    :raise ValidationError: If the input doesn't have the current ``item`` in a recursive non-leaf-input-object lookup
    :return: The :func:`__getitem__()` return value, or an :class:`lxml.etree.Element` attribute
    """

    def __init__(
        self,
        item: Any | tuple,
        default: Any = None,
        strict: bool = False,
    ):
        self.item = item
        self.default = default
        self.strict = strict


class UnionSchema(SchemaContainer):
    """
    Validate multiple schemas on the same input.

    Example:

    .. code-block:: python

        schema = validate.Schema(
            validate.union((
                validate.transform(str.format, one="abc", two="def"),
                validate.transform(str.format, one="123", two="456"),
            )),
        )
        assert schema.validate("{one} {two}") == ("abc def", "123 456")

    .. code-block:: python

        schema = validate.Schema(
            validate.union({
                "one": lambda val: val < 3,
                validate.optional("two"): lambda val: val > 1,
            }),
        )
        assert schema.validate(1) == {"one": 1}
        assert schema.validate(2) == {"one": 2, "two": 2}
        schema.validate(3)  # raises ValidationError

    :param Union[tuple, list, set, frozenset, dict] schema: A :class:`tuple`, :class:`list`, :class:`set`, :class:`frozenset`
                                                            or :class:`dict` of schemas
    :raises ValidationError: If a sequence item or the value of a non-optional key-value pair doesn't validate
    :return: A new object of the same type, with each item or key-value pair being validated against the same input value
    """


class UnionGetSchema:
    """
    Validate multiple :class:`GetItemSchema` schemas on the same input.

    Convenience wrapper for ``validate.union((validate.get(...), validate.get(...), ...))``.

    Example:

    .. code-block:: python

        schema = validate.Schema(
            validate.union_get("a", "b", ("c", "d")),
        )
        assert schema.validate({"a": 1, "b": 2, "c": {"d": 3}}) == (1, 2, 3)

    :param \\*getters: Inputs for each :class:`GetItemSchema`
    :return: A :class:`tuple` (default ``seq`` type) with items of the respective :class:`GetItemSchema` validations
    """

    def __init__(
        self,
        *getters,
        seq: type[tuple | list | set | frozenset] = tuple,
    ):
        self.getters: Sequence[GetItemSchema] = tuple(GetItemSchema(getter) for getter in getters)
        self.seq = seq


class RegexSchema:  # noqa: B903
    """
    A :class:`re.Pattern` that **must** match.

    Allows selecting a different regex pattern method (default is :meth:`re.Pattern.search()`).

    Please note the difference between :class:`RegexSchema`
    and the :func:`re.Pattern <streamlink.plugin.api.validate.validate_pattern()>` validation.

    Example:

    .. code-block:: python

        schema = validate.Schema(
            validate.regex(re.compile(r"Hello, (?P<name>\\w+)!")),
        )
        assert schema.validate("Hello, world!")["name"] == "world"
        schema.validate("Does not match")  # raises ValidationError
        schema.validate(123)  # raises ValidationError
        schema.validate(b"Hello, world!")  # raises ValidationError

    .. code-block:: python

        schema = validate.Schema(
            validate.regex(re.compile(r"Hello, (?P<name>\\w+)!"), method="findall"),
        )
        assert schema.validate("Hello, userA! Hello, userB!") == ["userA", "userB"]
        assert schema.validate("Does not match") == []  # findall does not return None

    :param pattern: A compiled pattern (:func:`re.compile` return value)
    :param method: The pattern's method which will be called when validating
    :raise ValidationError: If the input is not an instance of ``str`` or ``bytes``
    :raise ValidationError: If the type of the input doesn't match the pattern's ``str``/``bytes`` type
    :raise ValidationError: If the return value of the chosen regex pattern method is ``None``
    """

    def __init__(
        self,
        pattern: Pattern,
        method: Literal["search", "match", "fullmatch", "findall", "split", "sub", "subn"] = "search",
    ):
        self.pattern = pattern
        self.method = method


class XmlElementSchema:  # noqa: B903
    """
    Validate an XML element.

    Example:

    .. code-block:: python

        schema = validate.Schema(
            validate.xml_element(
                tag="foo",
                attrib={"bar": str},
                text=validate.transform(str.upper),
            ),
        )
        elem = lxml.etree.XML(\"\"\"<foo bar="baz">qux</foo>\"\"\")
        new_elem = schema.validate(elem)
        assert new_elem is not elem
        assert new_elem.tag == "foo"
        assert new_elem.attrib == {"bar": "baz"}
        assert new_elem.text == "QUX"
        assert new_elem.tail is None
        schema.validate(123)  # raises ValidationError
        schema.validate(lxml.etree.XML(\"\"\"<unknown/>\"\"\"))  # raises ValidationError

    :param tag: Optional element tag validation
    :param text: Optional element text validation
    :param attrib: Optional element attributes validation
    :param tail: Optional element tail validation
    :raise ValidationError: If ``value`` is not an :class:`lxml.etree.Element`
    :return: A new :class:`lxml.etree.Element` object, including a deep-copy of the input's child nodes,
             with optionally validated ``tag``, ``attrib`` mapping, ``text`` or ``tail``.
    """

    # signature is weird because of backwards compatiblity
    def __init__(
        self,
        tag: Any = None,
        text: Any = None,
        attrib: Any = None,
        tail: Any = None,
    ):
        self.tag = tag
        self.attrib = attrib
        self.text = text
        self.tail = tail



"""
$description Bulgarian CDN hosting live content for various websites in Bulgaria.
$url armymedia.bg
$url bgonair.bg
$url bloombergtv.bg
$url bnt.bg
$url i.cdn.bg
$url nova.bg
$url mu-vi.tv
$type live
$region Bulgaria
"""

import re
from urllib.parse import urlparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.utils.url import update_scheme


log = getLogger(__name__)


@pluginmatcher(
    name="armymedia",
    pattern=re.compile(r"https?://(?:www\.)?armymedia\.bg/?"),
)
@pluginmatcher(
    name="bgonair",
    pattern=re.compile(r"https?://(?:www\.)?bgonair\.bg/tvonline/?"),
)
@pluginmatcher(
    name="bloombergtv",
    pattern=re.compile(r"https?://(?:www\.)?bloombergtv\.bg/video/?"),
)
@pluginmatcher(
    name="bnt",
    pattern=re.compile(r"https?://(?:www\.)?(?:tv\.)?bnt\.bg/\w+(?:/\w+)?/?"),
)
@pluginmatcher(
    name="nova",
    pattern=re.compile(r"https?://(?:www\.)?nova\.bg/live/?"),
)
@pluginmatcher(
    name="mu-vi",
    pattern=re.compile(r"https?://(?:www\.)?mu-vi\.tv/LiveStreams/pages/Live\.aspx/?"),
)
@pluginmatcher(
    name="cdnbg",
    pattern=re.compile(r"https?://(?:www\.)?i\.cdn\.bg/live/?"),
)
class CDNBG(Plugin):
    @staticmethod
    def _find_url_regex(regex: re.Pattern) -> validate.all:
        return validate.all(
            validate.regex(regex),
            validate.get("url"),
        )

    @staticmethod
    def _find_url_xpath(xpath: str) -> validate.all:
        return validate.all(
            validate.xml_xpath_string(xpath),
            str,
        )

    def _get_streams(self):
        if "cdn.bg" in urlparse(self.url).netloc:
            player_url = self.url
            h = self.session.get_option("http-headers")
            if not h or not h.get("Referer"):
                log.error('Missing Referer for iframe URL, use --http-header "Referer=URL" ')
                return
            referer = h.get("Referer")
        else:
            referer = self.url
            player_url = self.session.http.get(
                self.url,
                schema=validate.Schema(
                    validate.any(
                        self._find_url_regex(
                            re.compile(r"'src',\s*'(?P<url>https?://i\.cdn\.bg/live/\w+)'\);"),
                        ),
                        validate.all(
                            validate.parse_html(),
                            validate.any(
                                self._find_url_xpath(".//iframe[contains(@src,'//i.cdn.bg/')][1]/@src"),
                                self._find_url_xpath(".//iframe[contains(@nitro-lazy-src,'//i.cdn.bg/')][1]/@nitro-lazy-src"),
                                self._find_url_xpath(".//source[contains(@src,'//i.cdn.bg/')][1]/@src"),
                                self._find_url_xpath(".//a[contains(@href,'//i.cdn.bg/')][1]/@href"),
                            ),
                        ),
                        # silently fail
                        validate.transform(lambda *_: None),
                    ),
                ),
            )

        if not player_url:
            log.error("Could not find player URL")
            return

        player_url = update_scheme("https://", player_url, force=False)
        log.debug(f"Found player URL: {player_url}")

        stream_url = self.session.http.get(
            player_url,
            headers={"Referer": referer},
            schema=validate.Schema(
                validate.any(
                    self._find_url_regex(
                        re.compile(r"sdata\.src.*?=.*?(?P<q>[\"'])(?P<url>.*?)(?P=q)"),
                    ),
                    self._find_url_regex(
                        re.compile(r"(src|file)\s*:\s*(?P<q>[\"'])(?P<url>(https?:)?//.+?m3u8.*?)(?P=q)"),
                    ),
                    validate.all(
                        validate.regex(re.compile(r"""window\.APP_CONFIG\.stream\s*=\s*(?P<url>".+?");""")),
                        validate.get("url"),
                        validate.parse_json(),
                    ),
                    validate.all(
                        validate.parse_html(),
                        validate.any(
                            self._find_url_xpath(".//video[contains(@src,'m3u8')][1]/@src"),
                            self._find_url_xpath(".//source[contains(@src,'m3u8')][1]/@src"),
                        ),
                    ),
                    # GEOBLOCKED
                    self._find_url_regex(
                        re.compile(r"(?P<url>[^\"]+geoblock[^\"]+)"),
                    ),
                ),
            ),
        )
        if "geoblock" in stream_url:
            log.error("Geo-restricted content")
            return

        stream_url = update_scheme(player_url, stream_url)

        return (
            HLSStream.parse_variant_playlist(self.session, stream_url, headers={"Referer": "https://i.cdn.bg/"})
            or {"live": HLSStream(self.session, stream_url, headers={"Referer": "https://i.cdn.bg/"})}
        )  # fmt: skip


__plugin__ = CDNBG



"""
$description Live TV channels and video on-demand service from RTPA, a Spanish public broadcaster.
$url rtpa.es
$type live, vod
$metadata title
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


@pluginmatcher(re.compile(r"https?://(?:www\.)?rtpa\.es"))
class RTPA(Plugin):
    def _get_streams(self):
        hls_url, self.title = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.union((
                    validate.xml_xpath_string(".//video/source[@src][@type='application/x-mpegURL'][1]/@src"),
                    validate.xml_xpath_string(".//head/title[1]/text()"),
                )),
            ),
        )
        if not hls_url:
            return
        return HLSStream.parse_variant_playlist(self.session, hls_url, headers={"Referer": self.url})


__plugin__ = RTPA



"""
$description Live TV channels from RSI and SRF, operations of SRG SSR, a Swiss public broadcaster.
$url srf.ch
$url rsi.ch
$type live
$region Switzerland
"""

import re
from urllib.parse import parse_qsl, urlparse, urlunparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:live\.(rsi)\.ch/|(?:www\.)?(srf)\.ch/sport/resultcenter)"),
)
class Swisstxt(Plugin):
    api_url = "http://event.api.swisstxt.ch/v1/stream/{site}/byEventItemIdAndType/{id}/HLS"

    def get_stream_url(self, event_id):
        site = self.match.group(1) or self.match.group(2)
        api_url = self.api_url.format(id=event_id, site=site.upper())
        log.debug(f"Calling API: {api_url}")

        stream_url = self.session.http.get(api_url).text.strip("\"'")

        parsed = urlparse(stream_url)
        query = dict(parse_qsl(parsed.query))
        return urlunparse(parsed._replace(query="")), query

    def _get_streams(self):
        event_id = dict(parse_qsl(urlparse(self.url).query.lower())).get("eventid")
        if event_id is None:
            return

        stream_url, params = self.get_stream_url(event_id)
        return HLSStream.parse_variant_playlist(self.session, stream_url, params=params)


__plugin__ = Swisstxt



"""
$description America-based television network centred towards business and capital market programming.
$url bloomberg.com
$type live, vod
$metadata title
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    name="live",
    pattern=re.compile(r"https?://(?:www\.)?bloomberg\.com/live(?:/(?P<channel>[^/]+))?"),
)
@pluginmatcher(
    name="vod",
    pattern=re.compile(r"https?://(?:www\.)?bloomberg\.com/news/videos/[^/]+/[^/]+"),
)
class Bloomberg(Plugin):
    LIVE_API_URL = "https://cdn.gotraffic.net/projector/latest/assets/config/config.min.json?v=1"
    VOD_API_URL = "https://www.bloomberg.com/api/embed?id={0}"
    DEFAULT_CHANNEL = "us"

    def _get_live_streams(self, data, channel):
        schema_live_ids = validate.Schema(
            {
                "live": {
                    "channels": {
                        "byChannelId": {
                            channel: validate.all(
                                {"liveId": str},
                                validate.get("liveId"),
                            ),
                        },
                    },
                },
            },
            validate.get(("live", "channels", "byChannelId", channel)),
        )
        try:
            live_id = schema_live_ids.validate(data)
        except PluginError:
            log.error(f"Could not find liveId for channel '{channel}'")
            return

        log.debug(f"Found liveId: {live_id}")
        return self.session.http.get(
            self.LIVE_API_URL,
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "livestreams": {
                        live_id: {
                            validate.optional("cdns"): validate.all(
                                [
                                    {
                                        "streams": [
                                            {
                                                "url": validate.url(),
                                            },
                                        ],
                                    },
                                ],
                                validate.transform(lambda x: [urls["url"] for y in x for urls in y["streams"]]),
                            ),
                        },
                    },
                },
                validate.get(("livestreams", live_id, "cdns")),
            ),
        )

    def _get_vod_streams(self, data):
        schema_vod_list = validate.Schema(
            validate.any(
                validate.all(
                    {"video": {"videoStory": dict}},
                    validate.get(("video", "videoStory")),
                ),
                validate.all(
                    {"quicktakeVideo": {"videoStory": dict}},
                    validate.get(("quicktakeVideo", "videoStory")),
                ),
            ),
            {
                "video": {
                    "bmmrId": str,
                },
            },
            validate.get(("video", "bmmrId")),
        )
        schema_url = validate.all(
            {"url": validate.url()},
            validate.get("url"),
        )

        try:
            video_id = schema_vod_list.validate(data)
        except PluginError:
            log.error("Could not find videoId")
            return

        log.debug(f"Found videoId: {video_id}")
        vod_url = self.VOD_API_URL.format(video_id)
        secureStreams, streams, self.title = self.session.http.get(
            vod_url,
            schema=validate.Schema(
                validate.parse_json(),
                {
                    validate.optional("secureStreams"): [schema_url],
                    validate.optional("streams"): [schema_url],
                    "title": str,
                },
                validate.union_get("secureStreams", "streams", "title"),
            ),
        )

        return secureStreams or streams

    def _get_streams(self):
        self.session.http.headers.clear()
        self.session.http.headers["User-Agent"] = useragents.CHROME

        data = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//script[contains(text(),'window.__PRELOADED_STATE__')][1]/text()"),
                validate.none_or_all(
                    re.compile(r"\bwindow\.__PRELOADED_STATE__\s*=\s*(?P<json>{.+?})\s*;(?:\s|$)"),
                    validate.none_or_all(
                        validate.get("json"),
                        validate.parse_json(),
                    ),
                ),
            ),
        )
        if not data:
            log.error("Could not find JSON data. Invalid URL or bot protection...")
            return

        if self.matches["live"]:
            streams = self._get_live_streams(data, self.match["channel"] or self.DEFAULT_CHANNEL)
        else:
            streams = self._get_vod_streams(data)

        if streams:
            # just return the first stream
            return HLSStream.parse_variant_playlist(self.session, streams[0])


__plugin__ = Bloomberg



"""
$description Tennis tournaments organized by the Association of Tennis Professionals.
$url atptour.com/en/atp-challenger-tour/challenger-tv
$type live, vod
"""

import re

from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate


@pluginmatcher(
    re.compile(r"https?://(?:www\.)?atptour\.com/(?:en|es)/atp-challenger-tour/challenger-tv"),
)
class AtpChallengerTour(Plugin):
    def _get_streams(self):
        iframe_url = self.session.http.get(
            self.url,
            schema=validate.Schema(
                validate.parse_html(),
                validate.xml_xpath_string(".//iframe[starts-with(@id,'vimeoPlayer_')][@src][1]/@src"),
                validate.any(None, validate.url()),
            ),
        )
        if iframe_url:
            return self.session.streams(iframe_url)


__plugin__ = AtpChallengerTour



"""
$description Live TV channels and video on-demand service from the SBA, a Saudi, state-owned broadcaster.
$url aloula.sba.sa
$url aloula.sa
$type live, vod
$metadata id
$metadata author
$metadata category
$metadata title
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream


log = getLogger(__name__)


@pluginmatcher(
    name="live",
    pattern=re.compile(r"https?://(?:aloula\.sba\.sa|(?:www\.)?aloula\.sa)/(?:\w{2}/)?live/(?P<live_slug>[^/?&]+)"),
)
@pluginmatcher(
    name="vod",
    pattern=re.compile(r"https?://(?:aloula\.sba\.sa|(?:www\.)?aloula\.sa)/(?:\w{2}/)?episode/(?P<vod_id>\d+)"),
)
class Aloula(Plugin):
    _URL_API_CHANNELS = "https://aloula.faulio.com/api/v1/channels"
    _URL_API_CHANNELS_PLAYER = "https://aloula.faulio.com/api/v1.1/channels/{channel}/player"
    _URL_API_VIDEO = "https://aloula.faulio.com/api/v1/video/{vod_id}"
    _URL_API_VIDEO_PLAYER = "https://aloula.faulio.com/api/v1/video/{vod_id}/player"

    def get_live(self, live_slug):
        live_data = self.session.http.get(
            self._URL_API_CHANNELS,
            schema=validate.Schema(
                validate.parse_json(),
                [
                    {
                        "id": int,
                        "url": str,
                        "title": str,
                        "has_live": bool,
                        "has_vod": bool,
                    },
                ],
                validate.filter(lambda k: k["url"] == live_slug),
            ),
        )
        if not live_data:
            return

        live_data = live_data[0]
        log.trace("%r", live_data)

        if not live_data["has_live"]:
            log.error("Stream is not live")
            return

        self.id = live_data["id"]
        self.author = "SBA"
        self.title = live_data["title"]
        self.category = "Live"

        hls_url = self.session.http.get(
            self._URL_API_CHANNELS_PLAYER.format(channel=self.id),
            schema=validate.Schema(
                validate.parse_json(),
                {"streams": {"hls": validate.url()}},
                validate.get(("streams", "hls")),
            ),
        )
        return HLSStream.parse_variant_playlist(self.session, hls_url)

    def get_vod(self, vod_id):
        vod_data = self.session.http.get(
            self._URL_API_VIDEO.format(vod_id=vod_id),
            acceptable_status=(200, 401),
            schema=validate.Schema(
                validate.parse_json(),
                validate.any(
                    validate.all(
                        {
                            "blocks": [
                                {
                                    "id": str,
                                    "program_title": str,
                                    "title": str,
                                    "season_number": int,
                                    "episode": int,
                                },
                            ],
                        },
                        validate.get(("blocks", 0)),
                    ),
                    {"cms_error": str, "message": str},
                ),
            ),
        )

        log.trace("%r", vod_data)

        if "cms_error" in vod_data and vod_data["cms_error"] == "auth":
            log.error("This stream requires a login; specify appropriate Authorization and profile HTTP headers")
            return

        if "cms_error" in vod_data:
            log.error(f"API error: {vod_data['cms_error']} ({vod_data['message']})")
            return

        self.id = vod_data["id"]
        self.author = vod_data["program_title"]
        self.title = vod_data["title"]
        self.category = f"S{vod_data['season_number']}E{vod_data['episode']}"

        hls_url = self.session.http.get(
            self._URL_API_VIDEO_PLAYER.format(vod_id=vod_id),
            schema=validate.Schema(
                validate.parse_json(),
                {"settings": {"protocols": {"hls": validate.url()}}},
                validate.get(("settings", "protocols", "hls")),
            ),
        )
        return HLSStream.parse_variant_playlist(self.session, hls_url)

    def _get_streams(self):
        self.session.http.headers.update({
            "Origin": "https://www.aloula.sa",
            "Referer": "https://www.aloula.sa/",
        })

        if self.matches["live"]:
            return self.get_live(self.match["live_slug"])
        else:
            return self.get_vod(self.match["vod_id"])


__plugin__ = Aloula



"""
$description Chinese, global live-streaming platform run by Huya Live.
$url nimo.tv
$type live
$metadata author
$metadata category
$metadata title
"""

import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import useragents, validate
from streamlink.stream.http import HTTPStream


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"https?://(?:www\.|m\.)?nimo\.tv/(?P<username>.*)"),
)
class NimoTV(Plugin):
    data_url = "https://m.nimo.tv/{0}"

    video_qualities = {
        250: "240p",
        500: "360p",
        1000: "480p",
        2500: "720p",
        6000: "1080p",
    }

    _re_appid = re.compile(rb"appid=(\d+)")
    _re_domain = re.compile(rb"(https?:\/\/[A-Za-z]{2,3}.hls[A-Za-z\.\/]+)(?:V|&)")
    _re_id = re.compile(rb"id=([^|\\]+)")
    _re_tp = re.compile(rb"tp=(\d+)")
    _re_wsSecret = re.compile(rb"wsSecret=(\w+)")
    _re_wsTime = re.compile(rb"wsTime=(\w+)")

    def _get_streams(self):
        username = self.match.group("username")
        if not username:
            return

        data = self.session.http.get(
            self.data_url.format(username),
            headers={
                "User-Agent": useragents.ANDROID,
            },
            schema=validate.Schema(
                re.compile(r"<script>var G_roomBaseInfo = ({.*?});</script>"),
                validate.none_or_all(
                    validate.get(1),
                    validate.parse_json(),
                    {
                        "title": str,
                        "nickname": str,
                        "game": str,
                        "liveStreamStatus": int,
                        validate.optional("mStreamPkg"): str,
                    },
                ),
            ),
        )

        if data["liveStreamStatus"] == 0:
            log.info("This stream is currently offline")
            return

        mStreamPkg = data.get("mStreamPkg")
        if not mStreamPkg:
            log.debug("missing mStreamPkg")
            return

        mStreamPkg = bytes.fromhex(mStreamPkg)
        try:
            appid = self._re_appid.search(mStreamPkg).group(1).decode("utf-8")  # type: ignore[ty:unresolved-attribute]
            domain = self._re_domain.search(mStreamPkg).group(1).decode("utf-8")  # type: ignore[ty:unresolved-attribute]
            id_ = self._re_id.search(mStreamPkg).group(1).decode("utf-8")  # type: ignore[ty:unresolved-attribute]
            tp = self._re_tp.search(mStreamPkg).group(1).decode("utf-8")  # type: ignore[ty:unresolved-attribute]
            ws_secret = self._re_wsSecret.search(mStreamPkg).group(1).decode("utf-8")  # type: ignore[ty:unresolved-attribute]
            ws_time = self._re_wsTime.search(mStreamPkg).group(1).decode("utf-8")  # type: ignore[ty:unresolved-attribute]
        except AttributeError:
            log.error("invalid mStreamPkg")
            return

        params = {
            "appid": appid,
            "id": id_,
            "tp": tp,
            "wsSecret": ws_secret,
            "wsTime": ws_time,
            "u": "0",
            "t": "100",
            "needwm": 1,
        }
        url = f"{domain}{id_}.flv"
        url = url.replace("hls.nimo.tv", "flv.nimo.tv")
        log.debug(f"URL={url}")
        for k, v in self.video_qualities.items():
            params = params.copy()
            params["ratio"] = k
            if v == "1080p":
                params["needwm"] = 0
            elif v in ("720p", "480p", "360p"):
                params["sphd"] = 1

            log.trace("%s params=%r", v, params)
            # some qualities might not exist, but it will select a different lower quality
            yield v, HTTPStream(self.session, url, params=params)

        self.author = data["nickname"]
        self.category = data["game"]
        self.title = data["title"]


__plugin__ = NimoTV



"""
$description Live local TV channels and video on-demand service. OTT service from FilmOn.
$url filmon.com
$type live, vod
$notes Some VODs are mp4 which may not stream, use -o to download
"""

from __future__ import annotations

import re
import time
from typing import TYPE_CHECKING
from urllib.parse import urlparse, urlunparse

from streamlink.exceptions import PluginError, StreamError
from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.session.http import TLSSecLevel1Adapter
from streamlink.stream.hls import HLSStream, HLSStreamReader, HLSStreamWorker
from streamlink.stream.http import HTTPStream


if TYPE_CHECKING:
    from collections.abc import Iterator


log = getLogger(__name__)


class FilmOnHLSStreamWorker(HLSStreamWorker):
    stream: FilmOnHLS

    def _fetch_playlist(self):
        try:
            return super()._fetch_playlist()
        except StreamError as err:
            if hasattr(err, "err") and err.err.response.status_code in (403, 502):  # type: ignore
                self.stream.watch_timeout = 0
                self.playlist_reload_time = 0
                log.debug(f"Force-reloading the channel playlist on error: {err}")
            raise err


class FilmOnHLSStreamReader(HLSStreamReader):
    __worker__ = FilmOnHLSStreamWorker


class FilmOnHLS(HLSStream):
    __shortname__ = "hls-filmon"
    __reader__ = FilmOnHLSStreamReader

    def __init__(self, session, url: str, api: FilmOnAPI, channel=None, vod_id=None, quality="high", **args):
        if channel is None and vod_id is None:
            raise PluginError("Channel or vod_id must be set")

        super().__init__(session, url, **args)
        self.api = api
        self.channel = channel
        self.vod_id = vod_id
        self.quality = quality
        self._url = url
        self.watch_timeout = 0.0
        self._first_netloc = ""

    def _get_stream_data(self) -> Iterator[tuple[str, str, int]]:
        if self.channel:
            log.debug(f"Reloading FilmOn channel playlist: {self.channel}")
            yield from self.api.channel(self.channel)
        elif self.vod_id:
            log.debug(f"Reloading FilmOn VOD playlist: {self.vod_id}")
            yield from self.api.vod(self.vod_id)

    @property
    def url(self) -> str:
        if time.time() <= self.watch_timeout:
            return self._url

        # If the watch timeout has passed then refresh the playlist from the API
        for quality, url, timeout in self._get_stream_data():
            if quality == self.quality:
                self.watch_timeout = time.time() + timeout

                if not self.channel:
                    self._url = url
                else:
                    parsed = urlparse(url)
                    if not self._first_netloc:
                        # save the first netloc
                        self._first_netloc = parsed.netloc
                    # always use the first saved netloc
                    self._url = parsed._replace(netloc=self._first_netloc).geturl()

                return self._url

        raise TypeError("Stream has expired and cannot be translated to a URL")


class FilmOnAPI:
    channel_url = "https://www.filmon.com/ajax/getChannelInfo"
    vod_url = "https://vms-admin.filmon.com/api/video/movie?id={0}"

    ATTEMPTS = 5
    TIMEOUT = 0.75

    stream_schema = validate.all(
        {
            "quality": str,
            "url": validate.url(),
            "watch-timeout": int,
        },
        validate.union_get("quality", "url", "watch-timeout"),
    )

    def __init__(self, session):
        self.session = session

    def channel(self, channel) -> list[tuple[str, str, int]]:
        num = 1
        while True:
            # retry for 50X errors or validation errors at the same time
            try:
                return self.session.http.post(
                    self.channel_url,
                    data={
                        "channel_id": channel,
                        "quality": "low",
                    },
                    headers={"X-Requested-With": "XMLHttpRequest"},
                    schema=validate.Schema(
                        validate.parse_json(),
                        {"streams": [self.stream_schema]},
                        validate.get("streams"),
                    ),
                )
            except PluginError:
                log.debug(f"Received invalid or non-JSON data, attempt {num}/{self.ATTEMPTS}")
                if num >= self.ATTEMPTS:
                    raise
                num = num + 1
                time.sleep(self.TIMEOUT)

    def vod(self, vod_id) -> list[tuple[str, str, int]]:
        return self.session.http.get(
            self.vod_url.format(vod_id),
            schema=validate.Schema(
                validate.parse_json(),
                {"response": {"streams": {str: self.stream_schema}}},
                validate.get(("response", "streams")),
                validate.transform(lambda d: d.values()),
            ),
        )


@pluginmatcher(
    re.compile(
        r"""
            https?://(?:www\.)?filmon\.(?:tv|com)/
            (?:
                (?:
                    index/popout\?
                    |
                    (?:tv/)?channel/(?:export\?)?
                    |
                    tv/(?!channel/)
                    |
                    channel/
                    |
                    (?P<is_group>group/)
                )(?:channel_id=)?(?P<channel>[-_\w]+)
                |
                vod/view/(?P<vod_id>[^/?&]+)
            )
        """,
        re.VERBOSE,
    ),
)
class Filmon(Plugin):
    quality_weights = {
        "high": 720,
        "low": 480,
    }

    TIME_CHANNEL = 60 * 60 * 24 * 365

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        parsed = urlparse(self.url)
        if parsed.path.startswith("/channel/"):
            self.url = urlunparse(parsed._replace(path=parsed.path.replace("/channel/", "/tv/")))
        self.api = FilmOnAPI(self.session)

        adapter = TLSSecLevel1Adapter()
        self.session.http.mount("https://filmon.com", adapter)
        self.session.http.mount("https://www.filmon.com", adapter)
        self.session.http.mount("https://vms-admin.filmon.com/", adapter)

        self.session.options.set("hls-playlist-reload-time", "segment")

    @classmethod
    def stream_weight(cls, stream: str) -> tuple[float, str]:
        weight = cls.quality_weights.get(stream)
        if weight:
            return weight, "filmon"

        return super().stream_weight(stream)

    def _get_streams(self):
        channel = self.match.group("channel")
        vod_id = self.match.group("vod_id")
        is_group = self.match.group("is_group")

        # get cookies
        self.session.http.get(self.url)

        if vod_id:
            for quality, url, _timeout in self.api.vod(vod_id):
                if url.endswith(".m3u8"):
                    streams = HLSStream.parse_variant_playlist(self.session, url)
                    if streams:
                        yield from streams.items()
                        return
                    yield quality, HLSStream(self.session, url)
                elif url.endswith(".mp4"):
                    yield quality, HTTPStream(self.session, url)
        else:
            if not channel or channel.isdigit():
                id_ = channel
            else:
                id_ = self.cache.get(channel)
                if id_ is not None:
                    log.debug(f"Found cached channel ID: {id_}")
                else:
                    id_ = self.session.http.get(
                        self.url,
                        schema=validate.Schema(
                            re.compile(r"""channel_id\s*=\s*(?P<q>['"]?)(?P<value>\d+)(?P=q)"""),
                            validate.any(None, validate.get("value")),
                        ),
                    )
                    log.debug(f"Found channel ID: {id_}")
                    # do not cache a group url
                    if id_ and not is_group:
                        self.cache.set(channel, id_, expires=self.TIME_CHANNEL)

            if id_ is None:
                raise PluginError(f"Unable to find channel ID: {channel}")

            try:
                for quality, url, _timeout in self.api.channel(id_):
                    yield quality, FilmOnHLS(self.session, url, self.api, channel=id_, quality=quality)
            except Exception:
                if channel and not channel.isdigit():
                    self.cache.set(channel, None, expires=0)
                    log.debug(f"Reset cached channel: {channel}")
                raise


__plugin__ = Filmon



import re

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, pluginmatcher
from streamlink.plugin.plugin import LOW_PRIORITY, parse_params, stream_weight
from streamlink.stream.dash import DASHStream
from streamlink.utils.url import update_scheme


log = getLogger(__name__)


@pluginmatcher(
    re.compile(r"dash://(?P<url>\S+)(?:\s(?P<params>.+))?$"),
)
@pluginmatcher(
    priority=LOW_PRIORITY,
    pattern=re.compile(
        # URL with explicit scheme, or URL with implicit HTTPS scheme and a path
        r"(?P<url>[^/]+/\S+\.mpd(?:\?\S*)?)(?:\s(?P<params>.+))?$",
        re.IGNORECASE,
    ),
)
class MPEGDASH(Plugin):
    @classmethod
    def stream_weight(cls, stream: str) -> tuple[float, str]:
        match = re.match(r"^(?:(.*)\+)?(?:a(\d+)k)$", stream)
        if match and match.group(1) and match.group(2):
            weight, group = stream_weight(match.group(1))
            weight += int(match.group(2))
            return weight, group
        elif match and match.group(2):
            return stream_weight(f"{match.group(2)}k")
        else:
            return stream_weight(stream)

    def _get_streams(self):
        data = self.match.groupdict()
        url = update_scheme("https://", str(data.get("url", "")), force=False)
        params = parse_params(data.get("params"))
        log.debug(f"URL={url}; params={params}")

        return DASHStream.parse_manifest(self.session, url, **params)


__plugin__ = MPEGDASH



# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all modules.
#
# CDP version: v0.0.1510116

from __future__ import annotations

from abc import ABC, ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, TypeAlias, cast


if TYPE_CHECKING:
    from collections.abc import MutableMapping

    from typing_extensions import Self


T_JSON_DICT: TypeAlias = dict[str, Any]
_event_parsers: MutableMapping[str, type[CDPEvent]] = {}


class _CDPEventMetaBase(type):
    def __new__(
        cls,
        name,
        bases,
        namespace,
        event: str | None = None,
        **kwargs,
    ) -> _CDPEventMetaBase:
        obj = super().__new__(cls, name, bases, namespace, **kwargs)
        if event is not None:
            _event_parsers[event] = cast("type[CDPEvent]", obj)

        return obj


class _CDPEventMeta(_CDPEventMetaBase, ABCMeta):
    pass


class CDPEvent(ABC, metaclass=_CDPEventMeta):
    @classmethod
    @abstractmethod
    def from_json(cls, json: T_JSON_DICT) -> Self:  # pragma: no cover
        raise NotImplementedError


def parse_json_event(json: T_JSON_DICT) -> CDPEvent:
    """Parse a JSON dictionary into a CDP event."""
    return _event_parsers[json["method"]].from_json(json["params"])



"""
$description Global live-streaming and video on-demand hosting platform.
$url players.brightcove.net
$type live, vod
$metadata title
"""

import re
from urllib.parse import urlparse

from streamlink.logger import getLogger
from streamlink.plugin import Plugin, PluginError, pluginmatcher
from streamlink.plugin.api import validate
from streamlink.stream.hls import HLSStream
from streamlink.stream.http import HTTPStream
from streamlink.utils.parse import parse_qsd


log = getLogger(__name__)


class BrightcovePlayer:
    URL_PLAYER = "https://players.brightcove.net/{account_id}/{player_id}/index.html?videoId={video_id}"
    URL_API = "https://edge.api.brightcove.com/playback/v1/accounts/{account_id}/videos/{video_id}"

    def __init__(self, session, account_id, player_id="default_default"):
        self.session = session
        self.account_id = account_id
        self.player_id = player_id
        self.title = None
        log.debug(f"Creating player for account {account_id} (player_id={player_id})")

    def get_streams(self, video_id):
        log.debug(f"Finding streams for video: {video_id}")

        player_url = self.URL_PLAYER.format(
            account_id=self.account_id,
            player_id=self.player_id,
            video_id=video_id,
        )

        policy_key = self.session.http.get(
            player_url,
            params={"videoId": video_id},
            schema=validate.Schema(
                re.compile(r"""policyKey\s*:\s*(?P<q>['"])(?P<key>[\w-]+)(?P=q)"""),
                validate.any(None, validate.get("key")),
            ),
        )
        if not policy_key:
            raise PluginError("Could not find Brightcove policy key")
        log.debug(f"Found policy key: {policy_key}")

        self.session.http.headers.update({"Referer": player_url})
        sources, self.title = self.session.http.get(
            self.URL_API.format(account_id=self.account_id, video_id=video_id),
            headers={"Accept": f"application/json;pk={policy_key}"},
            schema=validate.Schema(
                validate.parse_json(),
                {
                    "sources": [
                        {
                            "src": validate.url(),
                            validate.optional("type"): str,
                            validate.optional("container"): str,
                            validate.optional("height"): int,
                            validate.optional("avg_bitrate"): int,
                        },
                    ],
                    validate.optional("name"): str,
                },
                validate.union_get("sources", "name"),
            ),
        )

        for source in sources:
            if source.get("type") in ("application/vnd.apple.mpegurl", "application/x-mpegURL"):
                yield from HLSStream.parse_variant_playlist(self.session, source.get("src")).items()

            elif source.get("container") == "MP4":
                # determine quality name
                if source.get("height"):
                    q = f"{source.get('height')}p"
                elif source.get("avg_bitrate"):
                    q = f"{source.get('avg_bitrate') // 1000}k"
                else:
                    q = "live"

                yield q, HTTPStream(self.session, source.get("src"))


@pluginmatcher(
    re.compile(r"https?://players\.brightcove\.net/(?P<account_id>[^/]+)/(?P<player_id>[^/]+)/index\.html"),
)
class Brightcove(Plugin):
    def _get_streams(self):
        video_id = parse_qsd(urlparse(self.url).query).get("videoId")
        player = BrightcovePlayer(self.session, self.match.group("account_id"), self.match.group("player_id"))
        streams = dict(player.get_streams(video_id))
        self.title = player.title

        return streams


__plugin__ = Brightcove



# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all modules.
#
# CDP version: v0.0.1510116
# CDP domain: Security

from __future__ import annotations

import enum
from collections.abc import Generator
from dataclasses import dataclass
from typing import Any

import streamlink.webbrowser.cdp.devtools.network as network
from streamlink.webbrowser.cdp.devtools.util import T_JSON_DICT, CDPEvent


class CertificateId(int):
    """
    An internal certificate ID value.
    """
    def to_json(self) -> int:
        return self

    @classmethod
    def from_json(cls, json: int) -> CertificateId:
        return cls(json)

    def __repr__(self):
        return f"CertificateId({super().__repr__()})"


class MixedContentType(enum.Enum):
    """
    A description of mixed content (HTTP resources on HTTPS pages), as defined by
    https://www.w3.org/TR/mixed-content/#categories
    """
    BLOCKABLE = "blockable"
    OPTIONALLY_BLOCKABLE = "optionally-blockable"
    NONE = "none"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> MixedContentType:
        return cls(json)


class SecurityState(enum.Enum):
    """
    The security level of a page or resource.
    """
    UNKNOWN = "unknown"
    NEUTRAL = "neutral"
    INSECURE = "insecure"
    SECURE = "secure"
    INFO = "info"
    INSECURE_BROKEN = "insecure-broken"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> SecurityState:
        return cls(json)


@dataclass
class CertificateSecurityState:
    """
    Details about the security state of the page certificate.
    """
    #: Protocol name (e.g. "TLS 1.2" or "QUIC").
    protocol: str

    #: Key Exchange used by the connection, or the empty string if not applicable.
    key_exchange: str

    #: Cipher name.
    cipher: str

    #: Page certificate.
    certificate: list[str]

    #: Certificate subject name.
    subject_name: str

    #: Name of the issuing CA.
    issuer: str

    #: Certificate valid from date.
    valid_from: network.TimeSinceEpoch

    #: Certificate valid to (expiration) date
    valid_to: network.TimeSinceEpoch

    #: True if the certificate uses a weak signature algorithm.
    certificate_has_weak_signature: bool

    #: True if the certificate has a SHA1 signature in the chain.
    certificate_has_sha1_signature: bool

    #: True if modern SSL
    modern_ssl: bool

    #: True if the connection is using an obsolete SSL protocol.
    obsolete_ssl_protocol: bool

    #: True if the connection is using an obsolete SSL key exchange.
    obsolete_ssl_key_exchange: bool

    #: True if the connection is using an obsolete SSL cipher.
    obsolete_ssl_cipher: bool

    #: True if the connection is using an obsolete SSL signature.
    obsolete_ssl_signature: bool

    #: (EC)DH group used by the connection, if applicable.
    key_exchange_group: str | None = None

    #: TLS MAC. Note that AEAD ciphers do not have separate MACs.
    mac: str | None = None

    #: The highest priority network error code, if the certificate has an error.
    certificate_network_error: str | None = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["protocol"] = self.protocol
        json["keyExchange"] = self.key_exchange
        json["cipher"] = self.cipher
        json["certificate"] = list(self.certificate)
        json["subjectName"] = self.subject_name
        json["issuer"] = self.issuer
        json["validFrom"] = self.valid_from.to_json()
        json["validTo"] = self.valid_to.to_json()
        json["certificateHasWeakSignature"] = self.certificate_has_weak_signature
        json["certificateHasSha1Signature"] = self.certificate_has_sha1_signature
        json["modernSSL"] = self.modern_ssl
        json["obsoleteSslProtocol"] = self.obsolete_ssl_protocol
        json["obsoleteSslKeyExchange"] = self.obsolete_ssl_key_exchange
        json["obsoleteSslCipher"] = self.obsolete_ssl_cipher
        json["obsoleteSslSignature"] = self.obsolete_ssl_signature
        if self.key_exchange_group is not None:
            json["keyExchangeGroup"] = self.key_exchange_group
        if self.mac is not None:
            json["mac"] = self.mac
        if self.certificate_network_error is not None:
            json["certificateNetworkError"] = self.certificate_network_error
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> CertificateSecurityState:
        return cls(
            protocol=str(json["protocol"]),
            key_exchange=str(json["keyExchange"]),
            cipher=str(json["cipher"]),
            certificate=[str(i) for i in json["certificate"]],
            subject_name=str(json["subjectName"]),
            issuer=str(json["issuer"]),
            valid_from=network.TimeSinceEpoch.from_json(json["validFrom"]),
            valid_to=network.TimeSinceEpoch.from_json(json["validTo"]),
            certificate_has_weak_signature=bool(json["certificateHasWeakSignature"]),
            certificate_has_sha1_signature=bool(json["certificateHasSha1Signature"]),
            modern_ssl=bool(json["modernSSL"]),
            obsolete_ssl_protocol=bool(json["obsoleteSslProtocol"]),
            obsolete_ssl_key_exchange=bool(json["obsoleteSslKeyExchange"]),
            obsolete_ssl_cipher=bool(json["obsoleteSslCipher"]),
            obsolete_ssl_signature=bool(json["obsoleteSslSignature"]),
            key_exchange_group=str(json["keyExchangeGroup"]) if "keyExchangeGroup" in json else None,
            mac=str(json["mac"]) if "mac" in json else None,
            certificate_network_error=str(json["certificateNetworkError"]) if "certificateNetworkError" in json else None,
        )


class SafetyTipStatus(enum.Enum):
    BAD_REPUTATION = "badReputation"
    LOOKALIKE = "lookalike"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> SafetyTipStatus:
        return cls(json)


@dataclass
class SafetyTipInfo:
    #: Describes whether the page triggers any safety tips or reputation warnings. Default is unknown.
    safety_tip_status: SafetyTipStatus

    #: The URL the safety tip suggested ("Did you mean?"). Only filled in for lookalike matches.
    safe_url: str | None = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["safetyTipStatus"] = self.safety_tip_status.to_json()
        if self.safe_url is not None:
            json["safeUrl"] = self.safe_url
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> SafetyTipInfo:
        return cls(
            safety_tip_status=SafetyTipStatus.from_json(json["safetyTipStatus"]),
            safe_url=str(json["safeUrl"]) if "safeUrl" in json else None,
        )


@dataclass
class VisibleSecurityState:
    """
    Security state information about the page.
    """
    #: The security level of the page.
    security_state: SecurityState

    #: Array of security state issues ids.
    security_state_issue_ids: list[str]

    #: Security state details about the page certificate.
    certificate_security_state: CertificateSecurityState | None = None

    #: The type of Safety Tip triggered on the page. Note that this field will be set even if the Safety Tip UI was not actually shown.
    safety_tip_info: SafetyTipInfo | None = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["securityState"] = self.security_state.to_json()
        json["securityStateIssueIds"] = list(self.security_state_issue_ids)
        if self.certificate_security_state is not None:
            json["certificateSecurityState"] = self.certificate_security_state.to_json()
        if self.safety_tip_info is not None:
            json["safetyTipInfo"] = self.safety_tip_info.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> VisibleSecurityState:
        return cls(
            security_state=SecurityState.from_json(json["securityState"]),
            security_state_issue_ids=[str(i) for i in json["securityStateIssueIds"]],
            certificate_security_state=CertificateSecurityState.from_json(json["certificateSecurityState"]) if "certificateSecurityState" in json else None,
            safety_tip_info=SafetyTipInfo.from_json(json["safetyTipInfo"]) if "safetyTipInfo" in json else None,
        )


@dataclass
class SecurityStateExplanation:
    """
    An explanation of an factor contributing to the security state.
    """
    #: Security state representing the severity of the factor being explained.
    security_state: SecurityState

    #: Title describing the type of factor.
    title: str

    #: Short phrase describing the type of factor.
    summary: str

    #: Full text explanation of the factor.
    description: str

    #: The type of mixed content described by the explanation.
    mixed_content_type: MixedContentType

    #: Page certificate.
    certificate: list[str]

    #: Recommendations to fix any issues.
    recommendations: list[str] | None = None

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["securityState"] = self.security_state.to_json()
        json["title"] = self.title
        json["summary"] = self.summary
        json["description"] = self.description
        json["mixedContentType"] = self.mixed_content_type.to_json()
        json["certificate"] = list(self.certificate)
        if self.recommendations is not None:
            json["recommendations"] = list(self.recommendations)
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> SecurityStateExplanation:
        return cls(
            security_state=SecurityState.from_json(json["securityState"]),
            title=str(json["title"]),
            summary=str(json["summary"]),
            description=str(json["description"]),
            mixed_content_type=MixedContentType.from_json(json["mixedContentType"]),
            certificate=[str(i) for i in json["certificate"]],
            recommendations=[str(i) for i in json["recommendations"]] if "recommendations" in json else None,
        )


@dataclass
class InsecureContentStatus:
    """
    Information about insecure content on the page.
    """
    #: Always false.
    ran_mixed_content: bool

    #: Always false.
    displayed_mixed_content: bool

    #: Always false.
    contained_mixed_form: bool

    #: Always false.
    ran_content_with_cert_errors: bool

    #: Always false.
    displayed_content_with_cert_errors: bool

    #: Always set to unknown.
    ran_insecure_content_style: SecurityState

    #: Always set to unknown.
    displayed_insecure_content_style: SecurityState

    def to_json(self) -> T_JSON_DICT:
        json: T_JSON_DICT = {}
        json["ranMixedContent"] = self.ran_mixed_content
        json["displayedMixedContent"] = self.displayed_mixed_content
        json["containedMixedForm"] = self.contained_mixed_form
        json["ranContentWithCertErrors"] = self.ran_content_with_cert_errors
        json["displayedContentWithCertErrors"] = self.displayed_content_with_cert_errors
        json["ranInsecureContentStyle"] = self.ran_insecure_content_style.to_json()
        json["displayedInsecureContentStyle"] = self.displayed_insecure_content_style.to_json()
        return json

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> InsecureContentStatus:
        return cls(
            ran_mixed_content=bool(json["ranMixedContent"]),
            displayed_mixed_content=bool(json["displayedMixedContent"]),
            contained_mixed_form=bool(json["containedMixedForm"]),
            ran_content_with_cert_errors=bool(json["ranContentWithCertErrors"]),
            displayed_content_with_cert_errors=bool(json["displayedContentWithCertErrors"]),
            ran_insecure_content_style=SecurityState.from_json(json["ranInsecureContentStyle"]),
            displayed_insecure_content_style=SecurityState.from_json(json["displayedInsecureContentStyle"]),
        )


class CertificateErrorAction(enum.Enum):
    """
    The action to take when a certificate error occurs. continue will continue processing the
    request and cancel will cancel the request.
    """
    CONTINUE = "continue"
    CANCEL = "cancel"

    def to_json(self) -> str:
        return self.value

    @classmethod
    def from_json(cls, json: str) -> CertificateErrorAction:
        return cls(json)


def disable() -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Disables tracking security state changes.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Security.disable",
    }
    yield cmd_dict


def enable() -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Enables tracking security state changes.
    """
    cmd_dict: T_JSON_DICT = {
        "method": "Security.enable",
    }
    yield cmd_dict


def set_ignore_certificate_errors(
    ignore: bool,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Enable/disable whether all certificate errors should be ignored.

    :param ignore: If true, all certificate errors will be ignored.
    """
    params: T_JSON_DICT = {}
    params["ignore"] = ignore
    cmd_dict: T_JSON_DICT = {
        "method": "Security.setIgnoreCertificateErrors",
        "params": params,
    }
    yield cmd_dict


def handle_certificate_error(
    event_id: int,
    action: CertificateErrorAction,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Handles a certificate error that fired a certificateError event.

    :param event_id: The ID of the event.
    :param action: The action to take on the certificate error.
    """
    params: T_JSON_DICT = {}
    params["eventId"] = event_id
    params["action"] = action.to_json()
    cmd_dict: T_JSON_DICT = {
        "method": "Security.handleCertificateError",
        "params": params,
    }
    yield cmd_dict


def set_override_certificate_errors(
    override: bool,
) -> Generator[T_JSON_DICT, T_JSON_DICT, None]:
    """
    Enable/disable overriding certificate errors. If enabled, all certificate error events need to
    be handled by the DevTools client and should be answered with ``handleCertificateError`` commands.

    :param override: If true, certificate errors will be overridden.
    """
    params: T_JSON_DICT = {}
    params["override"] = override
    cmd_dict: T_JSON_DICT = {
        "method": "Security.setOverrideCertificateErrors",
        "params": params,
    }
    yield cmd_dict


@dataclass
class CertificateError(CDPEvent, event="Security.certificateError"):
    """
    There is a certificate error. If overriding certificate errors is enabled, then it should be
    handled with the ``handleCertificateError`` command. Note: this event does not fire if the
    certificate error has been allowed internally. Only one client per target should override
    certificate errors at the same time.
    """
    #: The ID of the event.
    event_id: int
    #: The type of the error.
    error_type: str
    #: The url that was requested.
    request_url: str

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> CertificateError:
        return cls(
            event_id=int(json["eventId"]),
            error_type=str(json["errorType"]),
            request_url=str(json["requestURL"]),
        )


@dataclass
class VisibleSecurityStateChanged(CDPEvent, event="Security.visibleSecurityStateChanged"):
    """
    **EXPERIMENTAL**

    The security state of the page changed.
    """
    #: Security state information about the page.
    visible_security_state: VisibleSecurityState

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> VisibleSecurityStateChanged:
        return cls(
            visible_security_state=VisibleSecurityState.from_json(json["visibleSecurityState"]),
        )


@dataclass
class SecurityStateChanged(CDPEvent, event="Security.securityStateChanged"):
    """
    The security state of the page changed. No longer being sent.
    """
    #: Security state.
    security_state: SecurityState
    #: True if the page was loaded over cryptographic transport such as HTTPS.
    scheme_is_cryptographic: bool
    #: Previously a list of explanations for the security state. Now always
    #: empty.
    explanations: list[SecurityStateExplanation]
    #: Information about insecure content on the page.
    insecure_content_status: InsecureContentStatus
    #: Overrides user-visible description of the state. Always omitted.
    summary: str | None

    @classmethod
    def from_json(cls, json: T_JSON_DICT) -> SecurityStateChanged:
        return cls(
            security_state=SecurityState.from_json(json["securityState"]),
            scheme_is_cryptographic=bool(json["schemeIsCryptographic"]),
            explanations=[SecurityStateExplanation.from_json(i) for i in json["explanations"]],
            insecure_content_status=InsecureContentStatus.from_json(json["insecureContentStatus"]),
            summary=str(json["summary"]) if "summary" in json else None,
        )



# DO NOT EDIT THIS FILE!
#
# This file is generated from the CDP specification. If you need to make
# changes, edit the generator and regenerate all modules.
#
# CDP version: v0.0.1510116

import streamlink.webbrowser.cdp.devtools.browser as browser
import streamlink.webbrowser.cdp.devtools.debugger as debugger
import streamlink.webbrowser.cdp.devtools.dom as dom
import streamlink.webbrowser.cdp.devtools.emulation as emulation
import streamlink.webbrowser.cdp.devtools.fetch as fetch
import streamlink.webbrowser.cdp.devtools.input_ as input_
import streamlink.webbrowser.cdp.devtools.inspector as inspector
import streamlink.webbrowser.cdp.devtools.io as io
import streamlink.webbrowser.cdp.devtools.network as network
import streamlink.webbrowser.cdp.devtools.page as page
import streamlink.webbrowser.cdp.devtools.runtime as runtime
import streamlink.webbrowser.cdp.devtools.security as security
import streamlink.webbrowser.cdp.devtools.target as target
import streamlink.webbrowser.cdp.devtools.util as util
